#include <cuda.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include "MersenneTwister.h"
using namespace std;

MTRand rand1;

__device__ double function(double x)
{
	return (x * sin(x*x) - (1-x) * sin(1-x*x));
}

double Hfunction(double x)
{
	return (x * sin(x*x) - (1-x) * sin(1-x*x));
}

__global__ void monteCarloStepUpper(double* inner_points, double* randXs, double* randYs)
{
	if(randYs[threadIdx.x] <= function(randXs[threadIdx.x]))
		inner_points[threadIdx.x] = 1;
}
__global__ void monteCarloStepLower(double* inner_points, double* randXs, double* randYs)
{
	if(randYs[threadIdx.x] >= function(randXs[threadIdx.x]))
		inner_points[threadIdx.x] = 1;
}

int main()
{
	const int monte_carlo_steps = 1024;
	double x_boundary = 1;
	double zero = 0.61360;
	double area_d_prime = (x_boundary-zero)*Hfunction(x_boundary); //Rectangle

	double* randXs = (double*)malloc(sizeof(double)*monte_carlo_steps);
	double* randYs = (double*)malloc(sizeof(double)*monte_carlo_steps);
	double* cudaRandXs;
	double* cudaRandYs;

	for(int i = 0; i < monte_carlo_steps; i++)
	{
		randXs[i] = zero+rand1.rand(x_boundary-zero);
		randYs[i] = rand1.rand(Hfunction(x_boundary));
	}

	double* inner_points = (double*)calloc(monte_carlo_steps, sizeof(double));
	double* cuda_inner_points;
	cudaMalloc(&cuda_inner_points, monte_carlo_steps*sizeof(double));
	cudaMemcpy(cuda_inner_points, inner_points, monte_carlo_steps*sizeof(double), cudaMemcpyHostToDevice);

	cudaMalloc(&cudaRandXs, sizeof(double)*monte_carlo_steps);
	cudaMalloc(&cudaRandYs, sizeof(double)*monte_carlo_steps);
	cudaMemcpy(cudaRandXs, randXs, sizeof(double)*monte_carlo_steps, cudaMemcpyHostToDevice);
	cudaMemcpy(cudaRandYs, randYs, sizeof(double)*monte_carlo_steps, cudaMemcpyHostToDevice);

	monteCarloStepUpper<<<1,monte_carlo_steps>>>(cuda_inner_points, cudaRandXs, cudaRandYs);

	cudaMemcpy(inner_points, cuda_inner_points, monte_carlo_steps*sizeof(double), cudaMemcpyDeviceToHost);
	long double I = 0;
	for(int q = 0; q < monte_carlo_steps; q++)
		if(inner_points[q] == 1)
			I++;

	long double estimate = (I/monte_carlo_steps) * area_d_prime;

	x_boundary = zero;
	area_d_prime = (x_boundary)*Hfunction(0); //Rectangle

	for(int i = 0; i < monte_carlo_steps; i++)
	{
		inner_points[i] = 0;
		randXs[i] = rand1.rand(x_boundary); //Random point within rectangle
		randYs[i] = rand1.rand(Hfunction(0));
	}
	cudaMemcpy(cudaRandXs, randXs, sizeof(double)*monte_carlo_steps, cudaMemcpyHostToDevice);
	cudaMemcpy(cudaRandYs, randYs, sizeof(double)*monte_carlo_steps, cudaMemcpyHostToDevice);
	cudaMemcpy(cuda_inner_points, inner_points, monte_carlo_steps*sizeof(double), cudaMemcpyHostToDevice);

	monteCarloStepLower<<<1,monte_carlo_steps>>>(cuda_inner_points, cudaRandXs, cudaRandYs);
	cudaMemcpy(inner_points, cuda_inner_points, monte_carlo_steps*sizeof(double), cudaMemcpyDeviceToHost);
	I = 0;
	for(int q = 0; q < monte_carlo_steps; q++)
		if(inner_points[q] == 1)
			I++;

	estimate += (I/monte_carlo_steps) * area_d_prime;

	double actual_value = -0.133795;
	cout << "Monte-Carlo Estimate: " << estimate << endl;
	cout << "Error: " << estimate - actual_value << endl;

	free(inner_points);
	cudaFree(cuda_inner_points);
	cudaFree(cudaRandXs);
	cudaFree(cudaRandYs);
	free(randXs);
	free(randYs);
	
	cin.get();
	return 0;
}